@@ -45,7 +45,11 @@ class QAbstractItemModel;
 QT_END_NAMESPACE
 
 namespace Core { class IOptionsPage; }
-namespace Utils { class MacroExpander; }
+
+namespace Utils {
+class MacroExpander;
+class ProcessHandle;
+} // Utils
 
 namespace Debugger {
 
@@ -196,6 +200,22 @@ public:
     const DebuggerRunParameters &runParameters() const;
     DebuggerRunParameters &runParameters();
 
+    enum {
+        // Remove need to qualify each use.
+        NeedsTemporaryStop = DebuggerCommand::NeedsTemporaryStop,
+        NeedsFullStop = DebuggerCommand::NeedsFullStop,
+        Discardable = DebuggerCommand::Discardable,
+        ConsoleCommand = DebuggerCommand::ConsoleCommand,
+        NeedsFlush = DebuggerCommand::NeedsFlush,
+        ExitRequest = DebuggerCommand::ExitRequest,
+        RunRequest = DebuggerCommand::RunRequest,
+        LosesChild = DebuggerCommand::LosesChild,
+        RebuildBreakpointModel = DebuggerCommand::RebuildBreakpointModel,
+        InUpdateLocals = DebuggerCommand::InUpdateLocals,
+        NativeCommand = DebuggerCommand::NativeCommand,
+        Silent = DebuggerCommand::Silent
+    };
+
     virtual bool canHandleToolTip(const DebuggerToolTipContext &) const;
     virtual void expandItem(const QString &iname); // Called when item in tree gets expanded.
     virtual void updateItem(const QString &iname); // Called for fresh watch items.
@@ -205,7 +225,7 @@ public:
     virtual void startDebugger(DebuggerRunControl *runControl);
     virtual void prepareForRestart() {}
 
-    virtual void watchPoint(const QPoint &);
+    virtual void watchPoint(const QPoint &pnt);
     virtual void runCommand(const DebuggerCommand &cmd);
     virtual void openMemoryView(const MemoryViewSetupData &data);
     virtual void fetchMemory(MemoryAgent *, quint64 addr, quint64 length);
@@ -284,7 +304,7 @@ public:
 
     static QString stateName(int s);
 
-    void notifyInferiorPid(qint64 pid);
+    void notifyInferiorPid(const Utils::ProcessHandle &pid);
     qint64 inferiorPid() const;
     bool isReverseDebugging() const;
     void handleCommand(int role, const QVariant &value);
@@ -314,6 +334,7 @@ public:
     void removeBreakpointMarker(const Breakpoint &bp);
 
     QString expand(const QString &string) const;
+    QString nativeStartupCommands() const;
 
 signals:
     void stateChanged(Debugger::DebuggerState state);
