@@ -1,7 +1,7 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing
+** Copyright (C) 2016 The Qt Company Ltd.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of Qt Creator.
 **
@@ -9,22 +9,17 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company.  For licensing terms and
-** conditions see http://www.qt.io/terms-conditions.  For further information
-** use the contact form at http://www.qt.io/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file.  Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, The Qt Company gives you certain additional
-** rights.  These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ****************************************************************************/
 
@@ -45,12 +40,12 @@
 #include "extraencodingsettings.h"
 #include "icodestylepreferences.h"
 #include "icodestylepreferencesfactory.h"
+#include "completionsettingspage.h"
 #include <texteditor/generichighlighter/highlightersettingspage.h>
 #include <texteditor/snippets/snippetssettingspage.h>
 
 #include <extensionsystem/pluginmanager.h>
 #include <coreplugin/icore.h>
-#include <utils/fadingindicator.h>
 #include <utils/qtcassert.h>
 
 #include <QApplication>
@@ -70,14 +65,13 @@ public:
     DisplaySettingsPage *m_displaySettingsPage;
     HighlighterSettingsPage *m_highlighterSettingsPage;
     SnippetsSettingsPage *m_snippetsSettingsPage;
+    CompletionSettingsPage *m_completionSettingsPage;
 
     QMap<Core::Id, ICodeStylePreferencesFactory *> m_languageToFactory;
 
     QMap<Core::Id, ICodeStylePreferences *> m_languageToCodeStyle;
     QMap<Core::Id, CodeStylePool *> m_languageToCodeStylePool;
     QMap<QString, Core::Id> m_mimeTypeToLanguage;
-
-    CompletionSettings m_completionSettings;
 };
 
 } // namespace Internal
@@ -124,6 +118,9 @@ TextEditorSettings::TextEditorSettings(QObject *parent)
     formatDescr.emplace_back(C_PARENTHESES_MISMATCH, tr("Mismatched Parentheses"),
                              tr("Displayed when mismatched parentheses, "
                                 "square brackets, or curly brackets are found."));
+    formatDescr.emplace_back(C_AUTOCOMPLETE, tr("Auto Complete"),
+                             tr("Displayed when a character is automatically inserted "
+                                "like brackets or quotes."));
     formatDescr.emplace_back(C_CURRENT_LINE, tr("Current Line"),
                              tr("Line where the cursor is placed in."),
                              FormatDescription::ShowBackgroundControl);
@@ -163,6 +160,7 @@ TextEditorSettings::TextEditorSettings(QObject *parent)
     formatDescr.emplace_back(C_LOCAL, tr("Local"), tr("Local variables."));
     formatDescr.emplace_back(C_FIELD, tr("Field"),
                              tr("Class' data members."), Qt::darkRed);
+    formatDescr.emplace_back(C_GLOBAL, tr("Global"), tr("Global variables."));
     formatDescr.emplace_back(C_ENUMERATION, tr("Enumeration"),
                              tr("Applied to enumeration items."), Qt::darkMagenta);
 
@@ -311,7 +309,14 @@ TextEditorSettings::TextEditorSettings(QObject *parent)
                              QColor(255, 190, 0),
                              QTextCharFormat::DotLine,
                              FormatDescription::ShowUnderlineControl);
-
+    formatDescr.emplace_back(C_DECLARATION,
+                             tr("Declaration"),
+                             tr("Declaration of a function, variable, and so on."),
+                             FormatDescription::ShowFontUnderlineAndRelativeControls);
+    formatDescr.emplace_back(C_OUTPUT_ARGUMENT,
+                             tr("Output Argument"),
+                             tr("Writable arguments of a function call."),
+                             FormatDescription::ShowFontUnderlineAndRelativeControls);
 
     d->m_fontSettingsPage = new FontSettingsPage(formatDescr,
                                                    Constants::TEXT_EDITOR_FONT_SETTINGS,
@@ -341,23 +346,27 @@ TextEditorSettings::TextEditorSettings(QObject *parent)
         new SnippetsSettingsPage(Constants::TEXT_EDITOR_SNIPPETS_SETTINGS, this);
     ExtensionSystem::PluginManager::addObject(d->m_snippetsSettingsPage);
 
-    connect(d->m_fontSettingsPage, SIGNAL(changed(TextEditor::FontSettings)),
-            this, SIGNAL(fontSettingsChanged(TextEditor::FontSettings)));
-    connect(d->m_behaviorSettingsPage, SIGNAL(typingSettingsChanged(TextEditor::TypingSettings)),
-            this, SIGNAL(typingSettingsChanged(TextEditor::TypingSettings)));
-    connect(d->m_behaviorSettingsPage, SIGNAL(storageSettingsChanged(TextEditor::StorageSettings)),
-            this, SIGNAL(storageSettingsChanged(TextEditor::StorageSettings)));
-    connect(d->m_behaviorSettingsPage, SIGNAL(behaviorSettingsChanged(TextEditor::BehaviorSettings)),
-            this, SIGNAL(behaviorSettingsChanged(TextEditor::BehaviorSettings)));
-    connect(d->m_behaviorSettingsPage, SIGNAL(extraEncodingSettingsChanged(TextEditor::ExtraEncodingSettings)),
-            this, SIGNAL(extraEncodingSettingsChanged(TextEditor::ExtraEncodingSettings)));
-    connect(d->m_displaySettingsPage, SIGNAL(marginSettingsChanged(TextEditor::MarginSettings)),
-            this, SIGNAL(marginSettingsChanged(TextEditor::MarginSettings)));
-    connect(d->m_displaySettingsPage, SIGNAL(displaySettingsChanged(TextEditor::DisplaySettings)),
-            this, SIGNAL(displaySettingsChanged(TextEditor::DisplaySettings)));
+    d->m_completionSettingsPage = new CompletionSettingsPage(this);
+    ExtensionSystem::PluginManager::addObject(d->m_completionSettingsPage);
 
-    // TODO: Move these settings to TextEditor category
-    d->m_completionSettings.fromSettings(QLatin1String("CppTools/"), Core::ICore::settings());
+    connect(d->m_fontSettingsPage, &FontSettingsPage::changed,
+            this, &TextEditorSettings::fontSettingsChanged);
+    connect(d->m_behaviorSettingsPage, &BehaviorSettingsPage::typingSettingsChanged,
+            this, &TextEditorSettings::typingSettingsChanged);
+    connect(d->m_behaviorSettingsPage, &BehaviorSettingsPage::storageSettingsChanged,
+            this, &TextEditorSettings::storageSettingsChanged);
+    connect(d->m_behaviorSettingsPage, &BehaviorSettingsPage::behaviorSettingsChanged,
+            this, &TextEditorSettings::behaviorSettingsChanged);
+    connect(d->m_behaviorSettingsPage, &BehaviorSettingsPage::extraEncodingSettingsChanged,
+            this, &TextEditorSettings::extraEncodingSettingsChanged);
+    connect(d->m_displaySettingsPage, &DisplaySettingsPage::marginSettingsChanged,
+            this, &TextEditorSettings::marginSettingsChanged);
+    connect(d->m_displaySettingsPage, &DisplaySettingsPage::displaySettingsChanged,
+            this, &TextEditorSettings::displaySettingsChanged);
+    connect(d->m_completionSettingsPage, &CompletionSettingsPage::completionSettingsChanged,
+            this, &TextEditorSettings::completionSettingsChanged);
+    connect(d->m_completionSettingsPage, &CompletionSettingsPage::commentsSettingsChanged,
+            this, &TextEditorSettings::commentsSettingsChanged);
 }
 
 TextEditorSettings::~TextEditorSettings()
@@ -367,6 +376,7 @@ TextEditorSettings::~TextEditorSettings()
     ExtensionSystem::PluginManager::removeObject(d->m_displaySettingsPage);
     ExtensionSystem::PluginManager::removeObject(d->m_highlighterSettingsPage);
     ExtensionSystem::PluginManager::removeObject(d->m_snippetsSettingsPage);
+    ExtensionSystem::PluginManager::removeObject(d->m_completionSettingsPage);
 
     delete d;
 
@@ -410,7 +420,7 @@ const DisplaySettings &TextEditorSettings::displaySettings()
 
 const CompletionSettings &TextEditorSettings::completionSettings()
 {
-    return d->m_completionSettings;
+    return d->m_completionSettingsPage->completionSettings();
 }
 
 const HighlighterSettings &TextEditorSettings::highlighterSettings()
@@ -423,15 +433,9 @@ const ExtraEncodingSettings &TextEditorSettings::extraEncodingSettings()
     return d->m_behaviorSettingsPage->extraEncodingSettings();
 }
 
-void TextEditorSettings::setCompletionSettings(const CompletionSettings &settings)
+const CommentsSettings &TextEditorSettings::commentsSettings()
 {
-    if (d->m_completionSettings == settings)
-        return;
-
-    d->m_completionSettings = settings;
-    d->m_completionSettings.toSettings(QLatin1String("CppTools/"), Core::ICore::settings());
-
-    emit m_instance->completionSettingsChanged(d->m_completionSettings);
+    return d->m_completionSettingsPage->commentsSettings();
 }
 
 void TextEditorSettings::registerCodeStyleFactory(ICodeStylePreferencesFactory *factory)
@@ -509,32 +513,21 @@ Core::Id TextEditorSettings::languageId(const QString &mimeType)
     return d->m_mimeTypeToLanguage.value(mimeType);
 }
 
-void showZoomIndicator(QWidget *editor, const int newZoom)
-{
-    if (editor) {
-        Utils::FadingIndicator::showText(editor,
-                                         QCoreApplication::translate("TextEditor::TextEditorSettings",
-                                                                     "Zoom: %1%").arg(newZoom),
-                                         Utils::FadingIndicator::SmallText);
-    }
-}
-
-void TextEditorSettings::fontZoomRequested(int zoom)
+int TextEditorSettings::increaseFontZoom(int step)
 {
     FontSettings &fs = const_cast<FontSettings&>(d->m_fontSettingsPage->fontSettings());
     const int previousZoom = fs.fontZoom();
-    const int newZoom = qMax(10, previousZoom + zoom);
-    if (newZoom == previousZoom)
-        return;
-    fs.setFontZoom(newZoom);
-    showZoomIndicator(qobject_cast<QWidget *>(sender()), newZoom);
-    d->m_fontSettingsPage->saveSettings();
+    const int newZoom = qMax(10, previousZoom + step);
+    if (newZoom != previousZoom) {
+        fs.setFontZoom(newZoom);
+        d->m_fontSettingsPage->saveSettings();
+    }
+    return newZoom;
 }
 
-void TextEditorSettings::zoomResetRequested()
+void TextEditorSettings::resetFontZoom()
 {
     FontSettings &fs = const_cast<FontSettings&>(d->m_fontSettingsPage->fontSettings());
     fs.setFontZoom(100);
-    showZoomIndicator(qobject_cast<QWidget *>(sender()), 100);
     d->m_fontSettingsPage->saveSettings();
 }
