@@ -1,7 +1,7 @@
-/**************************************************************************
+/****************************************************************************
 **
-** Copyright (c) 2014 Brian McGillion and Hugues Delorme
-** Contact: http://www.qt-project.org/legal
+** Copyright (C) 2016 Brian McGillion and Hugues Delorme
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of Qt Creator.
 **
@@ -9,32 +9,30 @@
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and Digia.  For licensing terms and
-** conditions see http://www.qt.io/licensing.  For further information
-** use the contact form at http://www.qt.io/contact-us.
+** a written agreement between you and The Qt Company. For licensing terms
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file.  Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Digia gives you certain additional
-** rights.  These rights are described in the Digia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 3 as published by the Free Software
+** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ****************************************************************************/
 
 #include "vcsbaseclient.h"
 #include "vcscommand.h"
 #include "vcsbaseclientsettings.h"
-#include "vcsbaseeditorparameterwidget.h"
+#include "vcsbaseeditorconfig.h"
 
 #include <coreplugin/icore.h>
 #include <coreplugin/vcsmanager.h>
+#include <coreplugin/editormanager/documentmodel.h>
+#include <coreplugin/editormanager/editormanager.h>
+#include <coreplugin/idocument.h>
 
 #include <utils/qtcassert.h>
 #include <utils/synchronousprocess.h>
@@ -45,7 +43,6 @@
 #include <QStringList>
 #include <QDir>
 #include <QProcess>
-#include <QSignalMapper>
 #include <QTextCodec>
 #include <QDebug>
 #include <QFileInfo>
@@ -64,8 +61,6 @@
     \sa VcsBase::VcsJobRunner
 */
 
-Q_DECLARE_METATYPE(QVariant)
-
 static Core::IEditor *locateEditor(const char *property, const QString &entry)
 {
     foreach (Core::IDocument *document, Core::DocumentModel::openedDocuments())
@@ -76,98 +71,239 @@ static Core::IEditor *locateEditor(const char *property, const QString &entry)
 
 namespace VcsBase {
 
+class VcsBaseClientImplPrivate
+{
+public:
+    VcsBaseClientImplPrivate(VcsBaseClientSettings *settings);
+    ~VcsBaseClientImplPrivate();
+
+    VcsBaseClientSettings *m_clientSettings;
+};
+
+VcsBaseClientImplPrivate::VcsBaseClientImplPrivate(VcsBaseClientSettings *settings) :
+    m_clientSettings(settings)
+{
+    m_clientSettings->readSettings(Core::ICore::settings());
+}
+
+VcsBaseClientImplPrivate::~VcsBaseClientImplPrivate()
+{
+    delete m_clientSettings;
+}
+
+VcsBaseClientImpl::VcsBaseClientImpl(VcsBaseClientSettings *settings) :
+    d(new VcsBaseClientImplPrivate(settings))
+{
+    connect(Core::ICore::instance(), &Core::ICore::saveSettingsRequested,
+            this, &VcsBaseClientImpl::saveSettings);
+}
+
+VcsBaseClientImpl::~VcsBaseClientImpl()
+{
+    delete d;
+}
+
+VcsBaseClientSettings &VcsBaseClientImpl::settings() const
+{
+    return *d->m_clientSettings;
+}
+
+Utils::FileName VcsBaseClientImpl::vcsBinary() const
+{
+    return settings().binaryPath();
+}
+
+VcsCommand *VcsBaseClientImpl::createCommand(const QString &workingDirectory,
+                                             VcsBaseEditorWidget *editor,
+                                             JobOutputBindMode mode) const
+{
+    auto cmd = new VcsCommand(workingDirectory, processEnvironment());
+    cmd->setDefaultTimeoutS(vcsTimeoutS());
+    if (editor)
+        editor->setCommand(cmd);
+    if (mode == VcsWindowOutputBind) {
+        cmd->addFlags(VcsCommand::ShowStdOut);
+        if (editor) // assume that the commands output is the important thing
+            cmd->addFlags(VcsCommand::SilentOutput);
+    } else if (editor) {
+        connect(cmd, &VcsCommand::stdOutText, editor, &VcsBaseEditorWidget::setPlainText);
+    }
+
+    return cmd;
+}
+
+void VcsBaseClientImpl::enqueueJob(VcsCommand *cmd, const QStringList &args,
+                                   const QString &workingDirectory,
+                                   const Utils::ExitCodeInterpreter &interpreter) const
+{
+    cmd->addJob(vcsBinary(), args, vcsTimeoutS(), workingDirectory, interpreter);
+    cmd->execute();
+}
+
+QProcessEnvironment VcsBaseClientImpl::processEnvironment() const
+{
+    QProcessEnvironment environment = QProcessEnvironment::systemEnvironment();
+    VcsBasePlugin::setProcessEnvironment(&environment, false);
+    return environment;
+}
+
+QString VcsBaseClientImpl::commandOutputFromLocal8Bit(const QByteArray &a)
+{
+    return Utils::SynchronousProcess::normalizeNewlines(QString::fromLocal8Bit(a));
+}
+
+QStringList VcsBaseClientImpl::commandOutputLinesFromLocal8Bit(const QByteArray &a)
+{
+    return splitLines(commandOutputFromLocal8Bit(a));
+}
+
+QStringList VcsBaseClientImpl::splitLines(const QString &s)
+{
+    const QChar newLine = QLatin1Char('\n');
+    QString output = s;
+    if (output.endsWith(newLine))
+        output.truncate(output.size() - 1);
+    if (output.isEmpty())
+        return QStringList();
+    return output.split(newLine);
+}
+
+QString VcsBaseClientImpl::stripLastNewline(const QString &in)
+{
+    if (in.endsWith('\n'))
+        return in.left(in.count() - 1);
+    return in;
+}
+
+Utils::SynchronousProcessResponse
+VcsBaseClientImpl::vcsFullySynchronousExec(const QString &workingDir, const Utils::FileName &binary,
+                                           const QStringList &args, unsigned flags,
+                                           int timeoutS, QTextCodec *codec) const
+{
+    VcsCommand command(workingDir, processEnvironment());
+    command.addFlags(flags);
+    if (codec)
+        command.setCodec(codec);
+    return command.runCommand(binary, args, (timeoutS > 0) ? timeoutS : vcsTimeoutS());
+}
+
+void VcsBaseClientImpl::resetCachedVcsInfo(const QString &workingDir)
+{
+    Core::VcsManager::resetVersionControlForDirectory(workingDir);
+}
+
+void VcsBaseClientImpl::annotateRevisionRequested(const QString &workingDirectory,
+                                                  const QString &file, const QString &change,
+                                                  int line)
+{
+    QString changeCopy = change;
+    // This might be invoked with a verbose revision description
+    // "SHA1 author subject" from the annotation context menu. Strip the rest.
+    const int blankPos = changeCopy.indexOf(QLatin1Char(' '));
+    if (blankPos != -1)
+        changeCopy.truncate(blankPos);
+    annotate(workingDirectory, file, changeCopy, line);
+}
+
+Utils::SynchronousProcessResponse
+VcsBaseClientImpl::vcsFullySynchronousExec(const QString &workingDir, const QStringList &args,
+                                           unsigned flags, int timeoutS, QTextCodec *codec) const
+{
+    return vcsFullySynchronousExec(workingDir, vcsBinary(), args, flags, timeoutS, codec);
+}
+
+VcsCommand *VcsBaseClientImpl::vcsExec(const QString &workingDirectory, const QStringList &arguments,
+                                       VcsBaseEditorWidget *editor, bool useOutputToWindow,
+                                       unsigned additionalFlags, const QVariant &cookie) const
+{
+    VcsCommand *command = createCommand(workingDirectory, editor,
+                                        useOutputToWindow ? VcsWindowOutputBind : NoOutputBind);
+    command->setCookie(cookie);
+    command->addFlags(additionalFlags);
+    if (editor)
+        command->setCodec(editor->codec());
+    enqueueJob(command, arguments);
+    return command;
+}
+
+Utils::SynchronousProcessResponse VcsBaseClientImpl::vcsSynchronousExec(const QString &workingDir,
+                                                                        const QStringList &args,
+                                                                        unsigned flags,
+                                                                        QTextCodec *outputCodec) const
+{
+    return VcsBasePlugin::runVcs(workingDir, vcsBinary(), args, vcsTimeoutS(), flags,
+                                 outputCodec, processEnvironment());
+}
+
+int VcsBaseClientImpl::vcsTimeoutS() const
+{
+    return settings().intValue(VcsBaseClientSettings::timeoutKey);
+}
+
+VcsBaseEditorWidget *VcsBaseClientImpl::createVcsEditor(Core::Id kind, QString title,
+                                                        const QString &source, QTextCodec *codec,
+                                                        const char *registerDynamicProperty,
+                                                        const QString &dynamicPropertyValue) const
+{
+    VcsBaseEditorWidget *baseEditor = 0;
+    Core::IEditor *outputEditor = locateEditor(registerDynamicProperty, dynamicPropertyValue);
+    const QString progressMsg = tr("Working...");
+    if (outputEditor) {
+        // Exists already
+        outputEditor->document()->setContents(progressMsg.toUtf8());
+        baseEditor = VcsBaseEditor::getVcsBaseEditor(outputEditor);
+        QTC_ASSERT(baseEditor, return 0);
+        Core::EditorManager::activateEditor(outputEditor);
+    } else {
+        outputEditor = Core::EditorManager::openEditorWithContents(kind, &title, progressMsg.toUtf8());
+        outputEditor->document()->setProperty(registerDynamicProperty, dynamicPropertyValue);
+        baseEditor = VcsBaseEditor::getVcsBaseEditor(outputEditor);
+        QTC_ASSERT(baseEditor, return 0);
+        connect(baseEditor, &VcsBaseEditorWidget::annotateRevisionRequested,
+                this, &VcsBaseClientImpl::annotateRevisionRequested);
+        baseEditor->setSource(source);
+        if (codec)
+            baseEditor->setCodec(codec);
+    }
+
+    baseEditor->setForceReadOnly(true);
+    return baseEditor;
+}
+
+void VcsBaseClientImpl::saveSettings()
+{
+    settings().writeSettings(Core::ICore::settings());
+}
+
 class VcsBaseClientPrivate
 {
 public:
-    VcsBaseClientPrivate(VcsBaseClient *client, VcsBaseClientSettings *settings);
+    VcsBaseEditorConfig *createDiffEditor(VcsBaseEditorWidget *editor);
+    VcsBaseEditorConfig *createLogEditor(VcsBaseEditorWidget *editor);
 
-    void statusParser(const QString &text);
-    void annotateRevision(const QString &workingDirectory, const QString &file,
-                          QString change, int lineNumber);
-    void saveSettings();
-
-    void bindCommandToEditor(VcsCommand *cmd, VcsBaseEditorWidget *editor);
-    void commandFinishedGotoLine(QWidget *editorObject);
-
-    VcsBaseClientSettings *m_clientSettings;
-    QSignalMapper *m_cmdFinishedMapper;
-
-private:
-    VcsBaseClient *m_client;
+    VcsBaseClient::ConfigCreator m_diffConfigCreator;
+    VcsBaseClient::ConfigCreator m_logConfigCreator;
 };
 
-VcsBaseClientPrivate::VcsBaseClientPrivate(VcsBaseClient *client, VcsBaseClientSettings *settings) :
-    m_clientSettings(settings),
-    m_cmdFinishedMapper(new QSignalMapper(client)),
-    m_client(client)
+VcsBaseEditorConfig *VcsBaseClientPrivate::createDiffEditor(VcsBaseEditorWidget *editor)
 {
+    return m_diffConfigCreator ? m_diffConfigCreator(editor->toolBar()) : 0;
 }
 
-void VcsBaseClientPrivate::statusParser(const QString &text)
+VcsBaseEditorConfig *VcsBaseClientPrivate::createLogEditor(VcsBaseEditorWidget *editor)
 {
-    QList<VcsBaseClient::StatusItem> lineInfoList;
-
-    QStringList rawStatusList = text.split(QLatin1Char('\n'));
-
-    foreach (const QString &string, rawStatusList) {
-        const VcsBaseClient::StatusItem lineInfo = m_client->parseStatusLine(string);
-        if (!lineInfo.flags.isEmpty() && !lineInfo.file.isEmpty())
-            lineInfoList.append(lineInfo);
-    }
-
-    emit m_client->parsedStatus(lineInfoList);
-}
-
-void VcsBaseClientPrivate::annotateRevision(const QString &workingDirectory,  const QString &file,
-                                            QString change, int lineNumber)
-{
-    // This might be invoked with a verbose revision description
-    // "SHA1 author subject" from the annotation context menu. Strip the rest.
-    const int blankPos = change.indexOf(QLatin1Char(' '));
-    if (blankPos != -1)
-        change.truncate(blankPos);
-    m_client->annotate(workingDirectory, file, change, lineNumber);
-}
-
-void VcsBaseClientPrivate::saveSettings()
-{
-    m_clientSettings->writeSettings(Core::ICore::settings());
-}
-
-void VcsBaseClientPrivate::bindCommandToEditor(VcsCommand *cmd, VcsBaseEditorWidget *editor)
-{
-    editor->setCommand(cmd);
-    QObject::connect(cmd, SIGNAL(finished(bool,int,QVariant)), m_cmdFinishedMapper, SLOT(map()));
-    m_cmdFinishedMapper->setMapping(cmd, editor);
-}
-
-void VcsBaseClientPrivate::commandFinishedGotoLine(QWidget *editorObject)
-{
-    VcsBaseEditorWidget *editor = qobject_cast<VcsBaseEditorWidget *>(editorObject);
-    VcsCommand *cmd = qobject_cast<VcsCommand *>(m_cmdFinishedMapper->mapping(editor));
-    if (editor && cmd) {
-        if (!cmd->lastExecutionSuccess()) {
-            editor->reportCommandFinished(false, cmd->lastExecutionExitCode(), cmd->cookie());
-        } else if (cmd->cookie().type() == QVariant::Int) {
-            const int line = cmd->cookie().toInt();
-            if (line >= 0)
-                editor->gotoLine(line);
-        }
-        m_cmdFinishedMapper->removeMappings(cmd);
-    }
+    return m_logConfigCreator ? m_logConfigCreator(editor->toolBar()) : 0;
 }
 
 VcsBaseClient::StatusItem::StatusItem(const QString &s, const QString &f) :
     flags(s), file(f)
-{
-}
+{ }
 
 VcsBaseClient::VcsBaseClient(VcsBaseClientSettings *settings) :
-    d(new VcsBaseClientPrivate(this, settings))
+    VcsBaseClientImpl(settings),
+    d(new VcsBaseClientPrivate)
 {
     qRegisterMetaType<QVariant>();
-    connect(Core::ICore::instance(), SIGNAL(saveSettingsRequested()), this, SLOT(saveSettings()));
-    connect(d->m_cmdFinishedMapper, SIGNAL(mapped(QWidget*)), this, SLOT(commandFinishedGotoLine(QWidget*)));
 }
 
 VcsBaseClient::~VcsBaseClient()
@@ -180,11 +316,10 @@ bool VcsBaseClient::synchronousCreateRepository(const QString &workingDirectory,
 {
     QStringList args(vcsCommandString(CreateRepositoryCommand));
     args << extraOptions;
-    QByteArray outputData;
-    if (!vcsFullySynchronousExec(workingDirectory, args, &outputData))
+    Utils::SynchronousProcessResponse result = vcsFullySynchronousExec(workingDirectory, args);
+    if (result.result != Utils::SynchronousProcessResponse::Finished)
         return false;
-    VcsOutputWindow::append(
-                Utils::SynchronousProcess::normalizeNewlines(QString::fromLocal8Bit(outputData)));
+    VcsOutputWindow::append(result.stdOut());
 
     resetCachedVcsInfo(workingDirectory);
 
@@ -199,10 +334,10 @@ bool VcsBaseClient::synchronousClone(const QString &workingDir,
     QStringList args;
     args << vcsCommandString(CloneCommand)
          << extraOptions << srcLocation << dstLocation;
-    QByteArray stdOut;
-    const bool cloneOk = vcsFullySynchronousExec(workingDir, args, &stdOut);
+
+    Utils::SynchronousProcessResponse result = vcsFullySynchronousExec(workingDir, args);
     resetCachedVcsInfo(workingDir);
-    return cloneOk;
+    return result.result == Utils::SynchronousProcessResponse::Finished;
 }
 
 bool VcsBaseClient::synchronousAdd(const QString &workingDir, const QString &filename,
@@ -210,8 +345,7 @@ bool VcsBaseClient::synchronousAdd(const QString &workingDir, const QString &fil
 {
     QStringList args;
     args << vcsCommandString(AddCommand) << extraOptions << filename;
-    QByteArray stdOut;
-    return vcsFullySynchronousExec(workingDir, args, &stdOut);
+    return vcsFullySynchronousExec(workingDir, args).result == Utils::SynchronousProcessResponse::Finished;
 }
 
 bool VcsBaseClient::synchronousRemove(const QString &workingDir, const QString &filename,
@@ -219,8 +353,7 @@ bool VcsBaseClient::synchronousRemove(const QString &workingDir, const QString &
 {
     QStringList args;
     args << vcsCommandString(RemoveCommand) << extraOptions << filename;
-    QByteArray stdOut;
-    return vcsFullySynchronousExec(workingDir, args, &stdOut);
+    return vcsFullySynchronousExec(workingDir, args).result == Utils::SynchronousProcessResponse::Finished;
 }
 
 bool VcsBaseClient::synchronousMove(const QString &workingDir,
@@ -229,8 +362,7 @@ bool VcsBaseClient::synchronousMove(const QString &workingDir,
 {
     QStringList args;
     args << vcsCommandString(MoveCommand) << extraOptions << from << to;
-    QByteArray stdOut;
-    return vcsFullySynchronousExec(workingDir, args, &stdOut);
+    return vcsFullySynchronousExec(workingDir, args).result == Utils::SynchronousProcessResponse::Finished;
 }
 
 bool VcsBaseClient::synchronousPull(const QString &workingDir,
@@ -241,9 +373,9 @@ bool VcsBaseClient::synchronousPull(const QString &workingDir,
     args << vcsCommandString(PullCommand) << extraOptions << srcLocation;
     // Disable UNIX terminals to suppress SSH prompting
     const unsigned flags =
-            VcsBasePlugin::SshPasswordPrompt
-            | VcsBasePlugin::ShowStdOutInLogWindow
-            | VcsBasePlugin::ShowSuccessMessage;
+            VcsCommand::SshPasswordPrompt
+            | VcsCommand::ShowStdOut
+            | VcsCommand::ShowSuccessMessage;
     const Utils::SynchronousProcessResponse resp = vcsSynchronousExec(workingDir, args, flags);
     const bool ok = resp.result == Utils::SynchronousProcessResponse::Finished;
     if (ok)
@@ -259,67 +391,16 @@ bool VcsBaseClient::synchronousPush(const QString &workingDir,
     args << vcsCommandString(PushCommand) << extraOptions << dstLocation;
     // Disable UNIX terminals to suppress SSH prompting
     const unsigned flags =
-            VcsBasePlugin::SshPasswordPrompt
-            | VcsBasePlugin::ShowStdOutInLogWindow
-            | VcsBasePlugin::ShowSuccessMessage;
+            VcsCommand::SshPasswordPrompt
+            | VcsCommand::ShowStdOut
+            | VcsCommand::ShowSuccessMessage;
     const Utils::SynchronousProcessResponse resp = vcsSynchronousExec(workingDir, args, flags);
     return resp.result == Utils::SynchronousProcessResponse::Finished;
 }
 
-bool VcsBaseClient::vcsFullySynchronousExec(const QString &workingDir,
-                                            const QStringList &args,
-                                            QByteArray *output) const
-{
-    QProcess vcsProcess;
-    if (!workingDir.isEmpty())
-        vcsProcess.setWorkingDirectory(workingDir);
-    vcsProcess.setProcessEnvironment(processEnvironment());
-
-    const Utils::FileName binary = settings()->binaryPath();
-
-    VcsOutputWindow::appendCommand(workingDir, binary, args);
-
-    vcsProcess.start(binary.toString(), args);
-
-    if (!vcsProcess.waitForStarted()) {
-        VcsOutputWindow::appendError(tr("Unable to start process \"%1\": %2")
-                                         .arg(binary.toUserOutput(), vcsProcess.errorString()));
-        return false;
-    }
-
-    vcsProcess.closeWriteChannel();
-
-    QByteArray stdErr;
-    const int timeoutSec = settings()->intValue(VcsBaseClientSettings::timeoutKey);
-    if (!Utils::SynchronousProcess::readDataFromProcess(vcsProcess, timeoutSec * 1000,
-                                                        output, &stdErr, true)) {
-        Utils::SynchronousProcess::stopProcess(vcsProcess);
-        VcsOutputWindow::appendError(tr("Timed out after %1s waiting for the process %2 to finish.")
-                                         .arg(timeoutSec).arg(binary.toUserOutput()));
-        return false;
-    }
-    if (!stdErr.isEmpty())
-        VcsOutputWindow::appendError(QString::fromLocal8Bit(stdErr));
-
-    return vcsProcess.exitStatus() == QProcess::NormalExit && vcsProcess.exitCode() == 0;
-}
-
-Utils::SynchronousProcessResponse VcsBaseClient::vcsSynchronousExec(
-        const QString &workingDirectory,
-        const QStringList &args,
-        unsigned flags,
-        QTextCodec *outputCodec) const
-{
-    const Utils::FileName binary = settings()->binaryPath();
-    const int timeoutSec = settings()->intValue(VcsBaseClientSettings::timeoutKey);
-    return VcsBasePlugin::runVcs(workingDirectory, binary, args,
-                                 timeoutSec * 1000, flags, outputCodec);
-}
-
-void VcsBaseClient::annotate(const QString &workingDir, const QString &file,
-                             const QString &revision /* = QString() */,
-                             int lineNumber /* = -1 */,
-                             const QStringList &extraOptions)
+VcsBaseEditorWidget *VcsBaseClient::annotate(
+        const QString &workingDir, const QString &file, const QString &revision /* = QString() */,
+        int lineNumber /* = -1 */, const QStringList &extraOptions)
 {
     const QString vcsCmdString = vcsCommandString(AnnotateCommand);
     QStringList args;
@@ -329,12 +410,14 @@ void VcsBaseClient::annotate(const QString &workingDir, const QString &file,
     const QString title = vcsEditorTitle(vcsCmdString, id);
     const QString source = VcsBaseEditor::getSource(workingDir, file);
 
-    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source, true,
+    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source,
+                                                  VcsBaseEditor::getCodec(source),
                                                   vcsCmdString.toLatin1().constData(), id);
 
     VcsCommand *cmd = createCommand(workingDir, editor);
     cmd->setCookie(lineNumber);
     enqueueJob(cmd, args);
+    return editor;
 }
 
 void VcsBaseClient::diff(const QString &workingDir, const QStringList &files,
@@ -345,25 +428,30 @@ void VcsBaseClient::diff(const QString &workingDir, const QStringList &files,
     const QString id = VcsBaseEditor::getTitleId(workingDir, files);
     const QString title = vcsEditorTitle(vcsCmdString, id);
     const QString source = VcsBaseEditor::getSource(workingDir, files);
-    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source, true,
+    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source,
+                                                  VcsBaseEditor::getCodec(source),
                                                   vcsCmdString.toLatin1().constData(), id);
     editor->setWorkingDirectory(workingDir);
 
-    VcsBaseEditorParameterWidget *paramWidget = editor->configurationWidget();
-    if (!paramWidget && (paramWidget = createDiffEditor(workingDir, files, extraOptions))) {
-        // editor has been just created, createVcsEditor() didn't set a configuration widget yet
-        connect(editor, &VcsBaseEditorWidget::diffChunkReverted,
-                paramWidget, &VcsBaseEditorParameterWidget::executeCommand);
-        editor->setConfigurationWidget(paramWidget);
+    QStringList effectiveArgs = extraOptions;
+    if (!editor->configurationAdded()) {
+        if (VcsBaseEditorConfig *paramWidget = d->createDiffEditor(editor)) {
+            // editor has been just created, createVcsEditor() didn't set a configuration widget yet
+            connect(editor, &VcsBaseEditorWidget::diffChunkReverted,
+                    paramWidget, &VcsBaseEditorConfig::executeCommand);
+            connect(paramWidget, &VcsBaseEditorConfig::commandExecutionRequested,
+                [=] { diff(workingDir, files, extraOptions + paramWidget->arguments()); } );
+            effectiveArgs += paramWidget->arguments();
+            editor->setConfigurationAdded();
+        }
     }
 
     QStringList args;
-    const QStringList paramArgs = paramWidget != 0 ? paramWidget->arguments() : QStringList();
-    args << vcsCmdString << extraOptions << paramArgs << files;
+    args << vcsCmdString << effectiveArgs << files;
     QTextCodec *codec = source.isEmpty() ? static_cast<QTextCodec *>(0) : VcsBaseEditor::getCodec(source);
     VcsCommand *command = createCommand(workingDir, editor);
     command->setCodec(codec);
-    enqueueJob(command, args, exitCodeInterpreter(DiffCommand, command));
+    enqueueJob(command, args, workingDir, exitCodeInterpreter(DiffCommand));
 }
 
 void VcsBaseClient::log(const QString &workingDir, const QStringList &files,
@@ -375,19 +463,25 @@ void VcsBaseClient::log(const QString &workingDir, const QStringList &files,
     const QString id = VcsBaseEditor::getTitleId(workingDir, files);
     const QString title = vcsEditorTitle(vcsCmdString, id);
     const QString source = VcsBaseEditor::getSource(workingDir, files);
-    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source, true,
+    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source,
+                                                  VcsBaseEditor::getCodec(source),
                                                   vcsCmdString.toLatin1().constData(), id);
     editor->setFileLogAnnotateEnabled(enableAnnotationContextMenu);
 
-    VcsBaseEditorParameterWidget *paramWidget = editor->configurationWidget();
-    if (!paramWidget && (paramWidget = createLogEditor(workingDir, files, extraOptions))) {
-        // editor has been just created, createVcsEditor() didn't set a configuration widget yet
-        editor->setConfigurationWidget(paramWidget);
+    QStringList effectiveArgs = extraOptions;
+    if (!editor->configurationAdded()) {
+        if (VcsBaseEditorConfig *paramWidget = d->createLogEditor(editor)) {
+            // editor has been just created, createVcsEditor() didn't set a configuration widget yet
+            connect(paramWidget, &VcsBaseEditorConfig::commandExecutionRequested,
+                [=]() { this->log(workingDir, files, extraOptions + paramWidget->arguments(),
+                                  enableAnnotationContextMenu); } );
+            effectiveArgs += paramWidget->arguments();
+            editor->setConfigurationAdded();
+        }
     }
 
     QStringList args;
-    const QStringList paramArgs = paramWidget != 0 ? paramWidget->arguments() : QStringList();
-    args << vcsCmdString << extraOptions << paramArgs << files;
+    args << vcsCmdString << effectiveArgs << files;
     enqueueJob(createCommand(workingDir, editor), args);
 }
 
@@ -401,7 +495,7 @@ void VcsBaseClient::revertFile(const QString &workingDir,
     // Indicate repository change or file list
     VcsCommand *cmd = createCommand(workingDir);
     cmd->setCookie(QStringList(workingDir + QLatin1Char('/') + file));
-    connect(cmd, SIGNAL(success(QVariant)), this, SIGNAL(changed(QVariant)), Qt::QueuedConnection);
+    connect(cmd, &VcsCommand::success, this, &VcsBaseClient::changed, Qt::QueuedConnection);
     enqueueJob(cmd, args);
 }
 
@@ -413,7 +507,7 @@ void VcsBaseClient::revertAll(const QString &workingDir, const QString &revision
     // Indicate repository change or file list
     VcsCommand *cmd = createCommand(workingDir);
     cmd->setCookie(QStringList(workingDir));
-    connect(cmd, SIGNAL(success(QVariant)), this, SIGNAL(changed(QVariant)), Qt::QueuedConnection);
+    connect(cmd, &VcsCommand::success, this, &VcsBaseClient::changed, Qt::QueuedConnection);
     enqueueJob(createCommand(workingDir), args);
 }
 
@@ -424,7 +518,8 @@ void VcsBaseClient::status(const QString &workingDir, const QString &file,
     args << extraOptions << file;
     VcsOutputWindow::setRepository(workingDir);
     VcsCommand *cmd = createCommand(workingDir, 0, VcsWindowOutputBind);
-    connect(cmd, SIGNAL(finished(bool,int,QVariant)), VcsOutputWindow::instance(), SLOT(clearRepository()),
+    connect(cmd, &VcsCommand::finished,
+            VcsOutputWindow::instance(), &VcsOutputWindow::clearRepository,
             Qt::QueuedConnection);
     enqueueJob(cmd, args);
 }
@@ -434,7 +529,7 @@ void VcsBaseClient::emitParsedStatus(const QString &repository, const QStringLis
     QStringList args(vcsCommandString(StatusCommand));
     args << extraOptions;
     VcsCommand *cmd = createCommand(repository);
-    connect(cmd, SIGNAL(output(QString)), this, SLOT(statusParser(QString)));
+    connect(cmd, &VcsCommand::stdOutText, this, &VcsBaseClient::statusParser);
     enqueueJob(cmd, args);
 }
 
@@ -460,11 +555,20 @@ QString VcsBaseClient::vcsCommandString(VcsCommandTag cmd) const
     return QString();
 }
 
-Utils::ExitCodeInterpreter *VcsBaseClient::exitCodeInterpreter(VcsCommandTag cmd, QObject *parent) const
+Utils::ExitCodeInterpreter VcsBaseClient::exitCodeInterpreter(VcsCommandTag cmd) const
 {
     Q_UNUSED(cmd)
-    Q_UNUSED(parent)
-    return 0;
+    return Utils::defaultExitCodeInterpreter;
+}
+
+void VcsBaseClient::setDiffConfigCreator(ConfigCreator creator)
+{
+    d->m_diffConfigCreator = std::move(creator);
+}
+
+void VcsBaseClient::setLogConfigCreator(ConfigCreator creator)
+{
+    d->m_logConfigCreator = std::move(creator);
 }
 
 void VcsBaseClient::import(const QString &repositoryRoot, const QStringList &files,
@@ -483,7 +587,8 @@ void VcsBaseClient::view(const QString &source, const QString &id,
     const Core::Id kind = vcsEditorKind(DiffCommand);
     const QString title = vcsEditorTitle(vcsCommandString(LogCommand), id);
 
-    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source, true, "view", id);
+    VcsBaseEditorWidget *editor = createVcsEditor(kind, title, source,
+                                                  VcsBaseEditor::getCodec(source), "view", id);
 
     const QFileInfo fi(source);
     const QString workingDirPath = fi.isFile() ? fi.absolutePath() : source;
@@ -497,7 +602,7 @@ void VcsBaseClient::update(const QString &repositoryRoot, const QString &revisio
     args << revisionSpec(revision) << extraOptions;
     VcsCommand *cmd = createCommand(repositoryRoot);
     cmd->setCookie(repositoryRoot);
-    connect(cmd, SIGNAL(success(QVariant)), this, SIGNAL(changed(QVariant)), Qt::QueuedConnection);
+    connect(cmd, &VcsCommand::success, this, &VcsBaseClient::changed, Qt::QueuedConnection);
     enqueueJob(cmd, args);
 }
 
@@ -514,111 +619,34 @@ void VcsBaseClient::commit(const QString &repositoryRoot,
     //   So descendants of VcsBaseClient *must* redefine commit() and extend
     //   extraOptions with the usage for commitMessageFile (see BazaarClient::commit()
     //   for example)
-    Q_UNUSED(commitMessageFile);
     QStringList args(vcsCommandString(CommitCommand));
     args << extraOptions << files;
-    enqueueJob(createCommand(repositoryRoot), args);
-}
-
-VcsBaseClientSettings *VcsBaseClient::settings() const
-{
-    return d->m_clientSettings;
-}
-
-VcsBaseEditorParameterWidget *VcsBaseClient::createDiffEditor(const QString &workingDir,
-                                                              const QStringList &files,
-                                                              const QStringList &extraOptions)
-{
-    Q_UNUSED(workingDir);
-    Q_UNUSED(files);
-    Q_UNUSED(extraOptions);
-    return 0;
-}
-
-VcsBaseEditorParameterWidget *VcsBaseClient::createLogEditor(const QString &workingDir,
-                                                             const QStringList &files,
-                                                             const QStringList &extraOptions)
-{
-    Q_UNUSED(workingDir);
-    Q_UNUSED(files);
-    Q_UNUSED(extraOptions);
-    return 0;
+    VcsCommand *cmd = createCommand(repositoryRoot, 0, VcsWindowOutputBind);
+    if (!commitMessageFile.isEmpty())
+        connect(cmd, &VcsCommand::finished, [commitMessageFile]() { QFile(commitMessageFile).remove(); });
+    enqueueJob(cmd, args);
 }
 
 QString VcsBaseClient::vcsEditorTitle(const QString &vcsCmd, const QString &sourceId) const
 {
-    const Utils::FileName binary = settings()->binaryPath();
-    return binary.toFileInfo().baseName() +
+    return vcsBinary().toFileInfo().baseName() +
             QLatin1Char(' ') + vcsCmd + QLatin1Char(' ') +
-            QFileInfo(sourceId).fileName();
+            Utils::FileName::fromString(sourceId).fileName();
 }
 
-VcsBaseEditorWidget *VcsBaseClient::createVcsEditor(Core::Id kind, QString title,
-                                                    const QString &source, bool setSourceCodec,
-                                                    const char *registerDynamicProperty,
-                                                    const QString &dynamicPropertyValue) const
+void VcsBaseClient::statusParser(const QString &text)
 {
-    VcsBaseEditorWidget *baseEditor = 0;
-    Core::IEditor *outputEditor = locateEditor(registerDynamicProperty, dynamicPropertyValue);
-    const QString progressMsg = tr("Working...");
-    if (outputEditor) {
-        // Exists already
-        outputEditor->document()->setContents(progressMsg.toUtf8());
-        baseEditor = VcsBaseEditor::getVcsBaseEditor(outputEditor);
-        QTC_ASSERT(baseEditor, return 0);
-        Core::EditorManager::activateEditor(outputEditor);
-    } else {
-        outputEditor = Core::EditorManager::openEditorWithContents(kind, &title, progressMsg.toUtf8());
-        outputEditor->document()->setProperty(registerDynamicProperty, dynamicPropertyValue);
-        baseEditor = VcsBaseEditor::getVcsBaseEditor(outputEditor);
-        connect(baseEditor, SIGNAL(annotateRevisionRequested(QString,QString,QString,int)),
-                this, SLOT(annotateRevision(QString,QString,QString,int)));
-        QTC_ASSERT(baseEditor, return 0);
-        baseEditor->setSource(source);
-        if (setSourceCodec)
-            baseEditor->setCodec(VcsBaseEditor::getCodec(source));
+    QList<VcsBaseClient::StatusItem> lineInfoList;
+
+    QStringList rawStatusList = text.split(QLatin1Char('\n'));
+
+    foreach (const QString &string, rawStatusList) {
+        const VcsBaseClient::StatusItem lineInfo = parseStatusLine(string);
+        if (!lineInfo.flags.isEmpty() && !lineInfo.file.isEmpty())
+            lineInfoList.append(lineInfo);
     }
 
-    baseEditor->setForceReadOnly(true);
-    return baseEditor;
-}
-
-QProcessEnvironment VcsBaseClient::processEnvironment() const
-{
-    QProcessEnvironment environment = QProcessEnvironment::systemEnvironment();
-    VcsBasePlugin::setProcessEnvironment(&environment, false);
-    return environment;
-}
-
-VcsCommand *VcsBaseClient::createCommand(const QString &workingDirectory,
-                                         VcsBaseEditorWidget *editor,
-                                         JobOutputBindMode mode) const
-{
-    VcsCommand *cmd = new VcsCommand(d->m_clientSettings->binaryPath(), workingDirectory,
-                                     processEnvironment());
-    cmd->setDefaultTimeout(d->m_clientSettings->intValue(VcsBaseClientSettings::timeoutKey));
-    if (editor)
-        d->bindCommandToEditor(cmd, editor);
-    if (mode == VcsWindowOutputBind) {
-        cmd->addFlags(VcsBasePlugin::ShowStdOutInLogWindow);
-        if (editor) // assume that the commands output is the important thing
-            cmd->addFlags(VcsBasePlugin::SilentOutput);
-    } else if (editor) {
-        connect(cmd, SIGNAL(output(QString)), editor, SLOT(setPlainText(QString)));
-    }
-
-    return cmd;
-}
-
-void VcsBaseClient::enqueueJob(VcsCommand *cmd, const QStringList &args, Utils::ExitCodeInterpreter *interpreter)
-{
-    cmd->addJob(args, interpreter);
-    cmd->execute();
-}
-
-void VcsBaseClient::resetCachedVcsInfo(const QString &workingDir)
-{
-    Core::VcsManager::resetVersionControlForDirectory(workingDir);
+    emit parsedStatus(lineInfoList);
 }
 
 } // namespace VcsBase
