@@ -40,15 +40,15 @@
 #include <qset.h>
 #include <qstringlist.h>
 #include <qtextstream.h>
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-# include <qjsondocument.h>
-# include <qjsonobject.h>
-# include <qjsonarray.h>
-#endif
+#include <qjsondocument.h>
+#include <qjsonobject.h>
+#include <qjsonarray.h>
 #ifdef PROEVALUATOR_THREAD_SAFE
 # include <qthreadpool.h>
 #endif
 
+#include <algorithm>
+
 #ifdef Q_OS_UNIX
 #include <time.h>
 #include <utime.h>
@@ -94,7 +94,7 @@ enum ExpandFunc {
 enum TestFunc {
     T_INVALID = 0, T_REQUIRES, T_GREATERTHAN, T_LESSTHAN, T_EQUALS,
     T_EXISTS, T_EXPORT, T_CLEAR, T_UNSET, T_EVAL, T_CONFIG, T_SYSTEM,
-    T_DEFINED, T_CONTAINS, T_INFILE,
+    T_DEFINED, T_DISCARD_FROM, T_CONTAINS, T_INFILE,
     T_COUNT, T_ISEMPTY, T_PARSE_JSON, T_INCLUDE, T_LOAD, T_DEBUG, T_LOG, T_MESSAGE, T_WARNING, T_ERROR, T_IF,
     T_MKPATH, T_WRITE_FILE, T_TOUCH, T_CACHE
 };
@@ -105,53 +105,53 @@ void QMakeEvaluator::initFunctionStatics()
         const char * const name;
         const ExpandFunc func;
     } expandInits[] = {
-        { "member", E_MEMBER },
-        { "str_member", E_STR_MEMBER },
-        { "first", E_FIRST },
-        { "take_first", E_TAKE_FIRST },
-        { "last", E_LAST },
-        { "take_last", E_TAKE_LAST },
-        { "size", E_SIZE },
-        { "str_size", E_STR_SIZE },
-        { "cat", E_CAT },
-        { "fromfile", E_FROMFILE },
-        { "eval", E_EVAL },
-        { "list", E_LIST },
-        { "sprintf", E_SPRINTF },
-        { "format_number", E_FORMAT_NUMBER },
-        { "num_add", E_NUM_ADD },
-        { "join", E_JOIN },
-        { "split", E_SPLIT },
-        { "basename", E_BASENAME },
-        { "dirname", E_DIRNAME },
-        { "section", E_SECTION },
-        { "find", E_FIND },
-        { "system", E_SYSTEM },
-        { "unique", E_UNIQUE },
-        { "sorted", E_SORTED },
-        { "reverse", E_REVERSE },
-        { "quote", E_QUOTE },
-        { "escape_expand", E_ESCAPE_EXPAND },
-        { "upper", E_UPPER },
-        { "lower", E_LOWER },
-        { "title", E_TITLE },
-        { "re_escape", E_RE_ESCAPE },
-        { "val_escape", E_VAL_ESCAPE },
-        { "files", E_FILES },
-        { "prompt", E_PROMPT },
-        { "replace", E_REPLACE },
-        { "sort_depends", E_SORT_DEPENDS },
-        { "resolve_depends", E_RESOLVE_DEPENDS },
-        { "enumerate_vars", E_ENUMERATE_VARS },
-        { "shadowed", E_SHADOWED },
-        { "absolute_path", E_ABSOLUTE_PATH },
-        { "relative_path", E_RELATIVE_PATH },
-        { "clean_path", E_CLEAN_PATH },
-        { "system_path", E_SYSTEM_PATH },
-        { "shell_path", E_SHELL_PATH },
-        { "system_quote", E_SYSTEM_QUOTE },
-        { "shell_quote", E_SHELL_QUOTE },
-        { "getenv", E_GETENV },
+        {"member", E_MEMBER},
+        {"str_member", E_STR_MEMBER},
+        {"first", E_FIRST},
+        {"take_first", E_TAKE_FIRST},
+        {"last", E_LAST},
+        {"take_last", E_TAKE_LAST},
+        {"size", E_SIZE},
+        {"str_size", E_STR_SIZE},
+        {"cat", E_CAT},
+        {"fromfile", E_FROMFILE},
+        {"eval", E_EVAL},
+        {"list", E_LIST},
+        {"sprintf", E_SPRINTF},
+        {"format_number", E_FORMAT_NUMBER},
+        {"num_add", E_NUM_ADD},
+        {"join", E_JOIN},
+        {"split", E_SPLIT},
+        {"basename", E_BASENAME},
+        {"dirname", E_DIRNAME},
+        {"section", E_SECTION},
+        {"find", E_FIND},
+        {"system", E_SYSTEM},
+        {"unique", E_UNIQUE},
+        {"sorted", E_SORTED},
+        {"reverse", E_REVERSE},
+        {"quote", E_QUOTE},
+        {"escape_expand", E_ESCAPE_EXPAND},
+        {"upper", E_UPPER},
+        {"lower", E_LOWER},
+        {"title", E_TITLE},
+        {"re_escape", E_RE_ESCAPE},
+        {"val_escape", E_VAL_ESCAPE},
+        {"files", E_FILES},
+        {"prompt", E_PROMPT},
+        {"replace", E_REPLACE},
+        {"sort_depends", E_SORT_DEPENDS},
+        {"resolve_depends", E_RESOLVE_DEPENDS},
+        {"enumerate_vars", E_ENUMERATE_VARS},
+        {"shadowed", E_SHADOWED},
+        {"absolute_path", E_ABSOLUTE_PATH},
+        {"relative_path", E_RELATIVE_PATH},
+        {"clean_path", E_CLEAN_PATH},
+        {"system_path", E_SYSTEM_PATH},
+        {"shell_path", E_SHELL_PATH},
+        {"system_quote", E_SYSTEM_QUOTE},
+        {"shell_quote", E_SHELL_QUOTE},
+        {"getenv", E_GETENV},
     };
     statics.expands.reserve((int)(sizeof(expandInits)/sizeof(expandInits[0])));
     for (unsigned i = 0; i < sizeof(expandInits)/sizeof(expandInits[0]); ++i)
@@ -161,39 +161,38 @@ void QMakeEvaluator::initFunctionStatics()
         const char * const name;
         const TestFunc func;
     } testInits[] = {
-        { "requires", T_REQUIRES },
-        { "greaterThan", T_GREATERTHAN },
-        { "lessThan", T_LESSTHAN },
-        { "equals", T_EQUALS },
-        { "isEqual", T_EQUALS },
-        { "exists", T_EXISTS },
-        { "export", T_EXPORT },
-        { "clear", T_CLEAR },
-        { "unset", T_UNSET },
-        { "eval", T_EVAL },
-        { "CONFIG", T_CONFIG },
-        { "if", T_IF },
-        { "isActiveConfig", T_CONFIG },
-        { "system", T_SYSTEM },
-        { "defined", T_DEFINED },
-        { "contains", T_CONTAINS },
-        { "infile", T_INFILE },
-        { "count", T_COUNT },
-        { "isEmpty", T_ISEMPTY },
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
-        { "parseJson", T_PARSE_JSON },
-#endif
-        { "load", T_LOAD },
-        { "include", T_INCLUDE },
-        { "debug", T_DEBUG },
-        { "log", T_LOG },
-        { "message", T_MESSAGE },
-        { "warning", T_WARNING },
-        { "error", T_ERROR },
-        { "mkpath", T_MKPATH },
-        { "write_file", T_WRITE_FILE },
-        { "touch", T_TOUCH },
-        { "cache", T_CACHE },
+        {"requires", T_REQUIRES},
+        {"greaterThan", T_GREATERTHAN},
+        {"lessThan", T_LESSTHAN},
+        {"equals", T_EQUALS},
+        {"isEqual", T_EQUALS},
+        {"exists", T_EXISTS},
+        {"export", T_EXPORT},
+        {"clear", T_CLEAR},
+        {"unset", T_UNSET},
+        {"eval", T_EVAL},
+        {"CONFIG", T_CONFIG},
+        {"if", T_IF},
+        {"isActiveConfig", T_CONFIG},
+        {"system", T_SYSTEM},
+        {"discard_from", T_DISCARD_FROM},
+        {"defined", T_DEFINED},
+        {"contains", T_CONTAINS},
+        {"infile", T_INFILE},
+        {"count", T_COUNT},
+        {"isEmpty", T_ISEMPTY},
+        {"parseJson", T_PARSE_JSON},
+        {"load", T_LOAD},
+        {"include", T_INCLUDE},
+        {"debug", T_DEBUG},
+        {"log", T_LOG},
+        {"message", T_MESSAGE},
+        {"warning", T_WARNING},
+        {"error", T_ERROR},
+        {"mkpath", T_MKPATH},
+        {"write_file", T_WRITE_FILE},
+        {"touch", T_TOUCH},
+        {"cache", T_CACHE},
     };
     statics.functions.reserve((int)(sizeof(testInits)/sizeof(testInits[0])));
     for (unsigned i = 0; i < sizeof(testInits)/sizeof(testInits[0]); ++i)
@@ -334,7 +333,6 @@ QMakeEvaluator::quoteValue(const ProString &val)
     return ret;
 }
 
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
 static void addJsonValue(const QJsonValue &value, const QString &keyPrefix, ProValueMap *map);
 
 static void insertJsonKeyValue(const QString &key, const QStringList &values, ProValueMap *map)
@@ -413,14 +411,13 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::parseJsonInto(const QByteArray &json
 
     return QMakeEvaluator::ReturnTrue;
 }
-#endif
 
 QMakeEvaluator::VisitReturn
 QMakeEvaluator::writeFile(const QString &ctx, const QString &fn, QIODevice::OpenMode mode,
-                          bool exe, const QString &contents)
+                          QMakeVfs::VfsFlags flags, const QString &contents)
 {
     QString errStr;
-    if (!m_vfs->writeFile(fn, mode, exe, contents, &errStr)) {
+    if (!m_vfs->writeFile(fn, mode, flags, contents, &errStr)) {
         evalError(fL1S("Cannot write %1file %2: %3")
                   .arg(ctx, QDir::toNativeSeparators(fn), errStr));
         return ReturnFalse;
@@ -447,12 +444,13 @@ void QMakeEvaluator::runProcess(QProcess *proc, const QString &command) const
 }
 #endif
 
-QByteArray QMakeEvaluator::getCommandOutput(const QString &args) const
+QByteArray QMakeEvaluator::getCommandOutput(const QString &args, int *exitCode) const
 {
     QByteArray out;
 #ifndef QT_BOOTSTRAPPED
     QProcess proc;
     runProcess(&proc, args);
+    *exitCode = (proc.exitStatus() == QProcess::NormalExit) ? proc.exitCode() : -1;
     QByteArray errout = proc.readAllStandardError();
 # ifdef PROEVALUATOR_FULL
     // FIXME: Qt really should have the option to set forwarding per channel
@@ -482,7 +480,12 @@ QByteArray QMakeEvaluator::getCommandOutput(const QString &args) const
                 break;
             out += QByteArray(buff, read_in);
         }
-        QT_PCLOSE(proc);
+        int ec = QT_PCLOSE(proc);
+# ifdef Q_OS_WIN
+        *exitCode = ec >= 0 ? ec : -1;
+# else
+        *exitCode = WIFEXITED(ec) ? WEXITSTATUS(ec) : -1;
+# endif
     }
 # ifdef Q_OS_WIN
     out.replace("\r\n", "\n");
@@ -591,7 +594,7 @@ ProStringList QMakeEvaluator::evaluateBuiltinExpand(
             int width = 0;
             bool zeropad = false;
             bool leftalign = false;
-            enum { DefaultSign, PadSign, AlwaysSign } sign = DefaultSign;
+            enum {DefaultSign, PadSign, AlwaysSign } sign = DefaultSign;
             if (args.count() >= 2) {
                 const auto opts = split_value_list(args.at(1).toQStringRef());
                 for (const ProString &opt : opts) {
@@ -872,8 +875,8 @@ ProStringList QMakeEvaluator::evaluateBuiltinExpand(
         break;
     case E_SYSTEM:
         if (!m_skipLevel) {
-            if (args.count() < 1 || args.count() > 2) {
-                evalError(fL1S("system(execute) requires one or two arguments."));
+            if (args.count() < 1 || args.count() > 3) {
+                evalError(fL1S("system(command, [mode], [stsvar]) requires one to three arguments."));
             } else {
                 bool blob = false;
                 bool lines = false;
@@ -887,7 +890,12 @@ ProStringList QMakeEvaluator::evaluateBuiltinExpand(
                     else if (!m_tmp2.compare(QLatin1String("lines"), Qt::CaseInsensitive))
                         lines = true;
                 }
-                QByteArray bytes = getCommandOutput(args.at(0).toQString(m_tmp2));
+                int exitCode;
+                QByteArray bytes = getCommandOutput(args.at(0).toQString(m_tmp2), &exitCode);
+                if (args.count() > 2 && !args.at(2).isEmpty()) {
+                    m_valuemapStack.top()[args.at(2).toKey()] =
+                            ProStringList(ProString(QString::number(exitCode)));
+                }
                 if (lines) {
                     QTextStream stream(bytes);
                     while (!stream.atEnd())
@@ -948,10 +956,10 @@ ProStringList QMakeEvaluator::evaluateBuiltinExpand(
                         struct {
                             char in, out;
                         } mapped_quotes[] = {
-                            { 'n', '\n' },
-                            { 't', '\t' },
-                            { 'r', '\r' },
-                            { 0, 0 }
+                            {'n', '\n' },
+                            {'t', '\t' },
+                            {'r', '\r' },
+                            {0, 0 }
                         };
                         for (int i = 0; mapped_quotes[i].in; ++i) {
                             if (*(i_data+x+1) == QLatin1Char(mapped_quotes[i].in)) {
@@ -1283,6 +1291,38 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
         }
         return ReturnTrue;
     }
+    case T_DISCARD_FROM: {
+        if (args.count() != 1 || args.at(0).isEmpty()) {
+            evalError(fL1S("discard_from(file) requires one argument."));
+            return ReturnFalse;
+        }
+        if (m_valuemapStack.count() != 1) {
+            evalError(fL1S("discard_from() cannot be called from functions."));
+            return ReturnFalse;
+        }
+        QString fn = resolvePath(args.at(0).toQString(m_tmp1));
+        ProFile *pro = m_parser->parsedProFile(fn, QMakeParser::ParseOnlyCached);
+        if (!pro)
+            return ReturnFalse;
+        ProValueMap &vmap = m_valuemapStack.first();
+        for (auto vit = vmap.begin(); vit != vmap.end(); ) {
+            if (!vit->isEmpty()) {
+                auto isFrom = [pro](const ProString &s) {
+                    return s.sourceFile() == pro;
+                };
+                vit->erase(std::remove_if(vit->begin(), vit->end(), isFrom), vit->end());
+                if (vit->isEmpty()) {
+                    // When an initially non-empty variable becomes entirely empty,
+                    // undefine it altogether.
+                    vit = vmap.erase(vit);
+                    continue;
+                }
+            }
+            ++vit;
+        }
+        pro->deref();
+        return ReturnTrue;
+    }
     case T_INFILE:
         if (args.count() < 2 || args.count() > 3) {
             evalError(fL1S("infile(file, var, [values]) requires two or three arguments."));
@@ -1488,7 +1528,6 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
             m_valuemapStack.top()[var] = statics.fakeValue;
         return ReturnTrue;
     }
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
     case T_PARSE_JSON: {
         if (args.count() != 2) {
             evalError(fL1S("parseJson(variable, into) requires two arguments."));
@@ -1499,7 +1538,6 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
         QString parseInto = args.at(1).toQString(m_tmp2);
         return parseJsonInto(json, parseInto, &m_valuemapStack.top());
     }
-#endif
     case T_INCLUDE: {
         if (args.count() < 1 || args.count() > 3) {
             evalError(fL1S("include(file, [into, [silent]]) requires one, two or three arguments."));
@@ -1589,7 +1627,7 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
 #ifdef PROEVALUATOR_FULL
                 fputs(msg.toLatin1().constData(), stderr);
 #endif
-            } else {
+            } else if (!msg.isEmpty() || func_t != T_ERROR) {
                 m_handler->fileMessage(
                         (func_t == T_ERROR   ? QMakeHandler::ErrorMessage :
                          func_t == T_WARNING ? QMakeHandler::WarningMessage :
@@ -1676,7 +1714,7 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
             return ReturnFalse;
         }
         QIODevice::OpenMode mode = QIODevice::Truncate;
-        bool exe = false;
+        QMakeVfs::VfsFlags flags = (m_cumulative ? QMakeVfs::VfsCumulative : QMakeVfs::VfsExact);
         QString contents;
         if (args.count() >= 2) {
             const ProStringList &vals = values(args.at(1).toKey());
@@ -1689,7 +1727,7 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
                     if (m_tmp3 == QLatin1String("append")) {
                         mode = QIODevice::Append;
                     } else if (m_tmp3 == QLatin1String("exe")) {
-                        exe = true;
+                        flags |= QMakeVfs::VfsExecutable;
                     } else {
                         evalError(fL1S("write_file(): invalid flag %1.").arg(m_tmp3));
                         return ReturnFalse;
@@ -1699,7 +1737,7 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
         }
         QString path = resolvePath(args.at(0).toQString(m_tmp1));
         path.detach(); // make sure to not leak m_tmp1 into the map of written files.
-        return writeFile(QString(), path, mode, exe, contents);
+        return writeFile(QString(), path, mode, flags, contents);
     }
     case T_TOUCH: {
         if (args.count() != 2) {
@@ -1883,6 +1921,7 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
             varstr += QLatin1Char('\n');
         }
         QString fn;
+        QMakeVfs::VfsFlags flags = (m_cumulative ? QMakeVfs::VfsCumulative : QMakeVfs::VfsExact);
         if (target == TargetSuper) {
             if (m_superfile.isEmpty()) {
                 m_superfile = QDir::cleanPath(m_outputDir + QLatin1String("/.qmake.super"));
@@ -1906,12 +1945,12 @@ QMakeEvaluator::VisitReturn QMakeEvaluator::evaluateBuiltinConditional(
             fn = m_stashfile;
             if (fn.isEmpty())
                 fn = QDir::cleanPath(m_outputDir + QLatin1String("/.qmake.stash"));
-            if (!m_vfs->exists(fn)) {
+            if (!m_vfs->exists(fn, flags)) {
                 printf("Info: creating stash file %s\n", qPrintable(QDir::toNativeSeparators(fn)));
                 valuesRef(ProKey("_QMAKE_STASH_")) << ProString(fn);
             }
         }
-        return writeFile(fL1S("cache "), fn, QIODevice::Append, false, varstr);
+        return writeFile(fL1S("cache "), fn, QIODevice::Append, flags, varstr);
     }
     default:
         evalError(fL1S("Function '%1' is not implemented.").arg(function.toQString(m_tmp1)));
